"""
pysat.constellations is a pysat module that provides
the interface for pysat to manage and analyze data
from multiple pysat instrument.  Each instrument
is contained within a subpackage of the pysat.instruments
package.
"""

__all__ = ['de2', 'icon']
